# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining BraketDeviceWrapper Class

"""
from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Union

from braket.ahs.analog_hamiltonian_simulation import AnalogHamiltonianSimulation
from braket.aws import AwsDevice
from braket.circuits import Circuit
from braket.circuits.measure import Measure

from qbraid.programs import NATIVE_REGISTRY, QPROGRAM_REGISTRY, ExperimentType, load_program
from qbraid.runtime.device import QuantumDevice
from qbraid.runtime.enums import DeviceStatus
from qbraid.runtime.exceptions import DeviceProgramTypeMismatchError
from qbraid.transpiler import transpile

from .availability import next_available_time
from .job import BraketQuantumTask

if TYPE_CHECKING:
    import braket.aws

    import qbraid.runtime
    import qbraid.runtime.aws
    import qbraid.transpiler


class BraketDevice(QuantumDevice):
    """Wrapper class for Amazon Braket ``Device`` objects."""

    def __init__(
        self,
        profile: qbraid.runtime.TargetProfile,
        session: Optional[braket.aws.AwsSession] = None,
    ):
        """Create a BraketDevice."""
        super().__init__(profile=profile)
        self._device = AwsDevice(arn=self.id, aws_session=session)
        self._provider_name = self.profile.get("provider_name")

    @property
    def name(self) -> str:
        """Return the name of this Device."""
        return self._device.name

    def __str__(self):
        """String representation of the BraketDevice object."""
        return f"{self.__class__.__name__}('{self._provider_name} {self.name}')"

    def status(self) -> qbraid.runtime.DeviceStatus:
        """Return the status of this Device."""
        if self._device.status == "ONLINE":
            if self._device.is_available:
                return DeviceStatus.ONLINE
            return DeviceStatus.UNAVAILABLE

        if self._device.status == "RETIRED":
            return DeviceStatus.RETIRED

        return DeviceStatus.OFFLINE

    def availability_window(self) -> tuple[bool, str, str]:
        """Provides device availability status. Indicates current availability,
        time remaining (hours, minutes, seconds) until next availability or
        unavailability, and future UTC datetime of next change in availability status.

        Returns:
            tuple[bool, str, Optional[str]]: Current device availability, hr/min/sec until
                availability switch, future UTC datetime of availability switch
        """
        return next_available_time(self._device)

    def queue_depth(self) -> int:
        """Return the number of jobs in the queue for the device."""
        queue_depth_info = self._device.queue_depth()
        total_queued = 0
        for queue_type in ["Normal", "Priority"]:
            num_queued = queue_depth_info.quantum_tasks[queue_type]
            if isinstance(num_queued, str) and num_queued.startswith(">"):
                num_queued = num_queued[1:]
            total_queued += int(num_queued)
        return total_queued

    def transform(
        self, run_input: Union[Circuit, AnalogHamiltonianSimulation]
    ) -> Union[Circuit, AnalogHamiltonianSimulation]:
        """Transpile a circuit for the device."""
        program = run_input

        provider = (self.profile.provider_name or "").upper()
        experiment_type = self.profile.experiment_type

        if experiment_type == ExperimentType.GATE_MODEL and not isinstance(program, Circuit):
            raise DeviceProgramTypeMismatchError(program, str(Circuit), experiment_type)

        if experiment_type == ExperimentType.AHS and not isinstance(
            program, AnalogHamiltonianSimulation
        ):
            raise DeviceProgramTypeMismatchError(
                program, str(AnalogHamiltonianSimulation), experiment_type
            )

        # Use Tket to transform circuits for IonQ to support a bigger gateset. The transformation
        # only applies when no measurement is presented in the circuit.
        includes_measurement = False
        if isinstance(run_input, Circuit):
            includes_measurement = any(
                isinstance(instruction.operator, Measure) for instruction in run_input.instructions
            )
        if provider == "IONQ" and isinstance(run_input, Circuit) and not includes_measurement:
            graph = self.scheme.conversion_graph
            if (
                graph is not None
                and graph.has_edge("pytket", "braket")
                and QPROGRAM_REGISTRY["pytket"] == NATIVE_REGISTRY["pytket"]
                and QPROGRAM_REGISTRY["braket"] == NATIVE_REGISTRY["braket"]
                and self._target_spec.alias == "braket"
            ):
                tk_circuit = transpile(program, "pytket", max_path_depth=1, conversion_graph=graph)
                tk_program = load_program(tk_circuit)
                tk_program.transform(self)
                tk_transformed = tk_program.program
                braket_transformed = transpile(
                    tk_transformed, "braket", max_path_depth=1, conversion_graph=graph
                )
                program = braket_transformed

        qprogram = load_program(program)
        qprogram.transform(self)
        program = qprogram.program

        return program

    def submit(
        self,
        run_input: Union[
            Circuit, AnalogHamiltonianSimulation, list[Circuit], list[AnalogHamiltonianSimulation]
        ],
        *args,
        **kwargs,
    ) -> Union[BraketQuantumTask, list[BraketQuantumTask]]:
        """Run a quantum task specification on this quantum device. Task must represent a
        quantum circuit, annealing problems not supported.

        Args:
            run_input: Specification of a task to run on device.

        Keyword Args:
            shots (int): The number of times to run the task on the device.

        Returns:
            The job like object for the run.

        """
        is_single_input = not isinstance(run_input, list)
        run_input = [run_input] if is_single_input else run_input

        if any(hasattr(circuit, "partial_measurement_qubits") for circuit in run_input):
            # Extract partial measurement qubit information and add as tags for job tracking
            tasks = []
            for circuit in run_input:
                tags: dict[str, str] = {}
                if hasattr(circuit, "partial_measurement_qubits"):
                    partial_measurement_qubits: list[int] = circuit.partial_measurement_qubits
                    # Convert qubit indices to a string format for tagging (e.g., "0/2/3")
                    tag_value = "/".join([str(x) for x in partial_measurement_qubits])
                    tags = {"partial_measurement_qubits": tag_value}
                task = self._device.run(circuit, tags=tags, *args, **kwargs)
                tasks.append(BraketQuantumTask(task.id, task=task, device=self._device))
        else:
            aws_quantum_task_batch = self._device.run_batch(run_input, *args, **kwargs)
            tasks = [
                BraketQuantumTask(task.id, task=task, device=self._device)
                for task in aws_quantum_task_batch.tasks
            ]

        if is_single_input:
            return tasks[0]
        return tasks



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint:disable=invalid-name

"""
Module defining IonQ job class

"""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Union

from qbraid_core._import import LazyLoader

from qbraid.runtime.enums import JobStatus
from qbraid.runtime.exceptions import QbraidRuntimeError
from qbraid.runtime.job import QuantumJob
from qbraid.runtime.postprocess import distribute_counts, normalize_data
from qbraid.runtime.result import Result
from qbraid.runtime.result_data import GateModelResultData, MeasCount, MeasProb

if TYPE_CHECKING:
    import qbraid.runtime.ionq

qbraid_rt_ionq: qbraid.runtime.ionq = LazyLoader("qbraid_rt_ionq", globals(), "qbraid.runtime.ionq")


class IonQJobError(QbraidRuntimeError):
    """Class for errors raised while processing an IonQ job."""


class IonQJob(QuantumJob):
    """IonQ job class."""

    def __init__(
        self, job_id: str, session: Optional[qbraid.runtime.ionq.IonQSession] = None, **kwargs
    ):
        super().__init__(job_id=job_id, **kwargs)
        self._session = session or qbraid_rt_ionq.IonQSession()

    @property
    def session(self) -> qbraid.runtime.ionq.IonQSession:
        """Return the IonQ session."""
        return self._session

    @staticmethod
    def _map_status(status: str) -> JobStatus:
        """Convert IonQ job status to qBraid job status."""
        status_map = {
            "submitted": JobStatus.INITIALIZING,
            "ready": JobStatus.QUEUED,
            "running": JobStatus.RUNNING,
            "failed": JobStatus.FAILED,
            "canceled": JobStatus.CANCELLED,
            "completed": JobStatus.COMPLETED,
        }
        return status_map.get(status, JobStatus.UNKNOWN)

    def status(self) -> JobStatus:
        """Return the current status of the IonQ job."""
        job_data = self.session.get_job(self.id)
        status = job_data.get("status")
        return self._map_status(status)

    def metadata(self) -> dict[str, Any]:
        """Store and return the metadata of the IonQ job."""
        if self.is_terminal_state() and "status" in self._cache_metadata:
            return self._cache_metadata

        job_metadata = self.session.get_job(self.id)
        self._cache_metadata.update(job_metadata)
        self._cache_metadata["status"] = self._map_status(self._cache_metadata["status"])
        return self._cache_metadata

    def cancel(self) -> None:
        """Cancel the IonQ job."""
        self.session.cancel_job(self.id)

    @staticmethod
    def _get_counts(result: dict[str, Any]) -> Union[MeasCount, list[MeasCount]]:
        """Return the raw counts of the run."""
        shots: Optional[int] = result.get("shots")
        probabilities: Optional[Union[MeasProb, dict[str, MeasProb]]] = result.get("probabilities")

        if shots is None or probabilities is None:
            raise ValueError("Missing shots or probabilities in result data.")

        def convert_to_counts(meas_prob: dict[str, float]) -> dict[str, int]:
            """Helper function to normalize probabilities and convert to counts."""
            probs_dec = {int(key): value for key, value in meas_prob.items()}
            probs_normal = normalize_data(probs_dec)
            return distribute_counts(probs_normal, shots)

        if all(isinstance(value, dict) for value in probabilities.values()):
            return [convert_to_counts(probs) for probs in probabilities.values()]

        return convert_to_counts(probabilities)

    def result(self) -> Result:
        """Return the result of the IonQ job."""
        self.wait_for_final_state()
        job_data = self.session.get_job(self.id)
        success = job_data.get("status") == "completed"
        if not success:
            failure: dict = job_data.get("failure", {})
            code = failure.get("code")
            message = failure.get("error")
            raise IonQJobError(f"Job failed with code {code}: {message}")

        results_url: str = job_data["results_url"]
        results_endpoint = results_url.split("v0.3")[-1]
        job_data["probabilities"] = self.session.get(results_endpoint).json()
        job_data["shots"] = job_data.get("shots", self._cache_metadata.get("shots"))

        measurement_counts = self._get_counts(job_data)
        data = GateModelResultData(measurement_counts=measurement_counts)
        return Result(
            device_id=job_data["target"], job_id=self.id, success=success, data=data, **job_data
        )



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining PennylaneTape Class

"""
from typing import TYPE_CHECKING

import pennylane as qml
from pennylane.tape import QuantumTape

from qbraid.programs.exceptions import ProgramTypeError

from ._model import GateModelProgram

if TYPE_CHECKING:
    import numpy as np


class PennylaneTape(GateModelProgram):
    """Wrapper class for Pennylane Quantum Tape programs."""

    def __init__(self, program: QuantumTape):
        super().__init__(program)
        if not isinstance(program, QuantumTape):
            raise ProgramTypeError(
                message=f"Expected 'pennylane.tape.QuantumTape' object, got '{type(program)}'."
            )

    @property
    def qubits(self) -> list[int]:
        """Returns the wires of the quantum tape as a list of ints"""
        return self.program.wires.tolist()

    @property
    def num_clbits(self) -> int:
        """Return the number of classical bits in the tape."""
        return 0

    @property
    def depth(self) -> int:
        """Calculates circuit depth of Pennylane tape"""
        tape = self.program

        op_count = {wire: 0 for wire in tape.wires}

        for op in tape.operations:
            for wire in op.wires:
                op_count[wire] += 1

        return max(op_count.values())

    def _unitary(self) -> "np.ndarray":
        """Return the unitary of the Pennylane tape"""
        return qml.matrix(self.program, wire_order=self.qubits)

    def remove_idle_qubits(self) -> None:
        """Applies a given wire map to all operations in a tape."""
        tape = self.program.copy()

        wires = sorted(tape.wires)
        contig_wires = list(range(len(wires)))

        wire_map = dict(zip(wires, contig_wires))

        [tape], _ = qml.map_wires(tape, wire_map)

        self._program = tape

    def reverse_qubit_order(self) -> None:
        """Reverses the wire ordering of a Pennylane tape."""
        tape = self.program.copy()

        wires_rev = sorted(tape.wires)
        wires = wires_rev.copy()
        wires_rev.reverse()

        wire_map = dict(zip(wires, wires_rev))

        [tape], _ = qml.map_wires(tape, wire_map)

        self._program = tape



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name,import-outside-toplevel

"""
Fixtures imported/defined in this file can be used by any test in this directory
without needing to import them (pytest will automatically discover them).

"""
import random
import textwrap
from typing import Optional

import numpy as np
import pytest

from qbraid.programs import NATIVE_REGISTRY, ExperimentType
from qbraid.runtime import TargetProfile
from qbraid.runtime.native import QbraidDevice, QbraidProvider
from qbraid.runtime.noise import NoiseModelSet
from qbraid.transpiler import Conversion, ConversionGraph, ConversionScheme

from ._resources import (
    DEVICE_DATA_AQUILA,
    DEVICE_DATA_QIR,
    DEVICE_DATA_QUERA_QASM,
    MockClient,
    MockDevice,
)


def _braket_circuit(meas=True):
    """Returns low-depth, one-qubit Braket circuit to be used for testing."""
    import braket.circuits

    circuit = braket.circuits.Circuit()
    circuit.h(0)
    circuit.ry(0, np.pi / 2)
    if meas:
        circuit.measure(0)
    return circuit


def _cirq_circuit(meas=True):
    """Returns Low-depth, one-qubit Cirq circuit to be used for testing.
    If ``meas=True``, applies measurement operation to end of circuit."""
    import cirq

    q0 = cirq.GridQubit(0, 0)

    def basic_circuit():
        yield cirq.H(q0)
        yield cirq.Ry(rads=np.pi / 2)(q0)
        if meas:
            yield cirq.measure(q0, key="q0")

    circuit = cirq.Circuit()
    circuit.append(basic_circuit())
    return circuit


def _qiskit_circuit(meas=True):
    """Returns Low-depth, one-qubit Qiskit circuit to be used for testing.
    If ``meas=True``, applies measurement operation to end of circuit."""
    import qiskit

    circuit = qiskit.QuantumCircuit(1, 1) if meas else qiskit.QuantumCircuit(1)
    circuit.h(0)
    circuit.ry(np.pi / 2, 0)
    if meas:
        circuit.measure(0, 0)
    return circuit


@pytest.fixture
def braket_circuit():
    """Returns low-depth, one-qubit Braket circuit to be used for testing."""
    return _braket_circuit()


@pytest.fixture
def qiskit_circuit():
    """Returns low-depth, one-qubit Qiskit circuit to be used for testing."""
    return _qiskit_circuit()


@pytest.fixture
def qasm3_circuit():
    """Returns low-depth, one-qubit QASM3 circuit to be used for testing."""
    qasm = """
    OPENQASM 3.0;
    bit[1] b;
    qubit[1] q;
    h q[0];
    ry(1.5707963267948966) q[0];
    b[0] = measure q[0];
    """
    return textwrap.dedent(qasm).strip()


@pytest.fixture
def run_inputs():
    """Returns list of test circuits for each available native provider."""
    circuits = []
    if "cirq" in NATIVE_REGISTRY:
        circuits.append(_cirq_circuit(meas=False))
    if "qiskit" in NATIVE_REGISTRY:
        circuits.append(_qiskit_circuit(meas=False))
    if "braket" in NATIVE_REGISTRY:
        circuits.append(_braket_circuit(meas=False))
    return circuits


@pytest.fixture
def circuit(request, run_inputs):
    """Return a circuit for testing."""
    index = request.param
    return run_inputs[index]


@pytest.fixture
def run_inputs_meas():
    """Returns list of test circuits for each available native provider."""
    circuits = []
    if "cirq" in NATIVE_REGISTRY:
        circuits.append(_cirq_circuit(meas=True))
    if "qiskit" in NATIVE_REGISTRY:
        circuits.append(_qiskit_circuit(meas=True))
    if "braket" in NATIVE_REGISTRY:
        circuits.append(_braket_circuit(meas=True))
    return circuits


@pytest.fixture
def circuit_meas(request, run_inputs_meas):
    """Return a circuit for testing."""
    index = request.param
    return run_inputs_meas[index]


@pytest.fixture
def device_data_qir():
    """Return a dictionary of device data for the qBraid QIR simulator."""
    return DEVICE_DATA_QIR


@pytest.fixture
def device_data_quera():
    """Return a dictionary of device data for the QuEra QASM simulator."""
    return DEVICE_DATA_QUERA_QASM


@pytest.fixture
def device_data_aquila():
    """Return a dictionary of device data for the QuEra Aquila QPU."""
    return DEVICE_DATA_AQUILA


@pytest.fixture
def valid_qasm2_no_meas() -> str:
    """Returns a valid qasm string."""
    qasm = """
    OPENQASM 2.0;
    include "qelib1.inc";
    qreg q[2];
    h q[0];
    cx q[0],q[1];
    cx q[1],q[0];
    """
    return textwrap.dedent(qasm).strip()


@pytest.fixture
def valid_qasm2():
    """Valid OpenQASM 2 string with measurement."""
    return """
    OPENQASM 2.0;
    include "qelib1.inc";
    qreg q[2];
    creg c0[1];
    creg c1[1];
    swap q[0],q[1];
    measure q[0] -> c0[0];
    measure q[1] -> c1[0];
    """


@pytest.fixture
def mock_client():
    """Mock client for testing."""
    return MockClient()


@pytest.fixture
def mock_provider(mock_client):
    """Mock provider for testing."""
    return QbraidProvider(client=mock_client)


@pytest.fixture
def mock_profile():
    """Mock profile for testing."""
    return TargetProfile(
        device_id="qbraid_qir_simulator",
        simulator=True,
        experiment_type=ExperimentType.GATE_MODEL,
        num_qubits=64,
        program_spec=QbraidProvider._get_program_spec("pyqir", "qbraid_qir_simulator"),
        noise_models=NoiseModelSet.from_iterable(["ideal"]),
    )


@pytest.fixture
def mock_scheme():
    """Mock conversion scheme for testing."""
    conv1 = Conversion("alice", "bob", lambda x: x + 1)
    conv2 = Conversion("bob", "charlie", lambda x: x - 1)
    graph = ConversionGraph(conversions=[conv1, conv2])
    scheme = ConversionScheme(conversion_graph=graph)
    return scheme


@pytest.fixture
def mock_qbraid_device(mock_profile, mock_scheme, mock_client):
    """Mock QbraidDevice for testing."""
    return QbraidDevice(profile=mock_profile, client=mock_client, scheme=mock_scheme)


@pytest.fixture
def mock_basic_device(mock_profile):
    """Generic mock device for testing."""
    return MockDevice(profile=mock_profile)


def _uniform_state_circuit_cirq(num_qubits: Optional[int] = None, measure: Optional[bool] = True):
    """
    Creates a Cirq circuit where all qubits are entangled to uniformly be in
    either |0⟩ or |1⟩ states upon measurement.

    This circuit initializes the first qubit in a superposition state using a
    Hadamard gate and then entangles all other qubits to this first qubit using
    CNOT gates. This ensures all qubits collapse to the same state upon measurement,
    resulting in either all |0⟩s or all |1⟩s uniformly across different executions.

    Args:
        num_qubits (optional, int): The number of qubits in the circuit. If not provided,
                                    a default random number between 10 and 20 is used.
        measure (optional, bool): Whether to measure the qubits at the end of the circuit.

    Returns:
        cirq.Circuit: The resulting circuit where the measurement outcome of all qubits is
                      either all |0⟩s or all |1⟩s.

    Raises:
        ValueError: If the number of qubits provided is less than 1.
    """
    import cirq

    if num_qubits is not None and num_qubits < 1:
        raise ValueError("Number of qubits must be at least 1.")

    num_qubits = num_qubits or random.randint(10, 20)

    qubits = [cirq.LineQubit(i) for i in range(num_qubits)]

    circuit = cirq.Circuit()

    circuit.append(cirq.H(qubits[0]))

    for qubit in qubits[1:]:
        circuit.append(cirq.CNOT(qubits[0], qubit))

    if measure:
        circuit.append(cirq.measure(*qubits, key="result"))

    return circuit


def _uniform_state_circuit_qiskit(num_qubits: Optional[int] = None, measure: Optional[bool] = True):
    """
    Creates a Qiskit circuit where all qubits are entangled to uniformly be in
    either |0⟩ or |1⟩ states upon measurement.

    Args:
        num_qubits (optional, int): The number of qubits in the circuit. If not provided,
            a default random number between 10 and 20 is used.
        measure (optional, bool): Whether to measure the qubits at the end of the circuit.

    Returns:
        QuantumCircuit: The resulting circuit where the measurement outcome of all qubits is
            either all |0⟩s or all |1⟩s.

    Raises:
        ValueError: If the number of qubits provided is less than 1.
    """
    from qiskit import QuantumCircuit

    if num_qubits is not None and num_qubits < 1:
        raise ValueError("Number of qubits must be at least 1.")

    num_qubits = num_qubits or random.randint(10, 20)

    circuit = QuantumCircuit(num_qubits)

    circuit.h(0)

    for qubit in range(1, num_qubits):
        circuit.cx(0, qubit)

    if measure:
        circuit.measure_all()

    return circuit


@pytest.fixture
def cirq_uniform():
    """Cirq circuit used for testing."""
    return _uniform_state_circuit_cirq


@pytest.fixture
def qiskit_uniform():
    """Qiskit circuit used for testing."""
    return _uniform_state_circuit_qiskit


@pytest.fixture
def braket_ahs():
    """Return an example AHS program."""
    import braket.ahs
    import braket.timings

    a = 5.7e-6  # meters

    register = braket.ahs.AtomArrangement()
    register.add(np.array([0.5, 0.5 + 1 / np.sqrt(2)]) * a)
    register.add(np.array([0.5 + 1 / np.sqrt(2), 0.5]) * a)
    register.add(np.array([0.5 + 1 / np.sqrt(2), -0.5]) * a)
    register.add(np.array([0.5, -0.5 - 1 / np.sqrt(2)]) * a)
    register.add(np.array([-0.5, -0.5 - 1 / np.sqrt(2)]) * a)
    register.add(np.array([-0.5 - 1 / np.sqrt(2), -0.5]) * a)
    register.add(np.array([-0.5 - 1 / np.sqrt(2), 0.5]) * a)
    register.add(np.array([-0.5, 0.5 + 1 / np.sqrt(2)]) * a)

    time_max = 4e-6  # seconds
    time_ramp = 1e-7  # seconds
    omega_max = 6300000.0  # rad / sec
    delta_start = -5 * omega_max
    delta_end = 5 * omega_max

    omega = braket.timings.TimeSeries()
    omega.put(0.0, 0.0)
    omega.put(time_ramp, omega_max)
    omega.put(time_max - time_ramp, omega_max)
    omega.put(time_max, 0.0)

    delta = braket.timings.TimeSeries()
    delta.put(0.0, delta_start)
    delta.put(time_ramp, delta_start)
    delta.put(time_max - time_ramp, delta_end)
    delta.put(time_max, delta_end)

    phi = braket.timings.TimeSeries().put(0.0, 0.0).put(time_max, 0.0)

    drive = braket.ahs.DrivingField(amplitude=omega, phase=phi, detuning=delta)

    return braket.ahs.AnalogHamiltonianSimulation(register=register, hamiltonian=drive)


@pytest.fixture
def ahs_dict():
    """Returns dictionary representation of AHS program."""
    return {
        "register": {
            "sites": [
                ["0.00000285", "0.00000688050865276332"],
                ["0.00000688050865276332", "0.00000285"],
                ["0.00000688050865276332", "-0.00000285"],
                ["0.00000285", "-0.00000688050865276332"],
                ["-0.00000285", "-0.00000688050865276332"],
                ["-0.00000688050865276332", "-0.00000285"],
                ["-0.00000688050865276332", "0.00000285"],
                ["-0.00000285", "0.00000688050865276332"],
            ],
            "filling": [1, 1, 1, 1, 1, 1, 1, 1],
        },
        "hamiltonian": {
            "drivingFields": [
                {
                    "amplitude": {
                        "time_series": {
                            "values": ["0.0", "6300000.0", "6300000.0", "0.0"],
                            "times": ["0.0", "1E-7", "0.0000039", "0.000004"],
                        },
                        "pattern": "uniform",
                    },
                    "phase": {
                        "time_series": {"values": ["0.0", "0.0"], "times": ["0.0", "0.000004"]},
                        "pattern": "uniform",
                    },
                    "detuning": {
                        "time_series": {
                            "values": ["-31500000.0", "-31500000.0", "31500000.0", "31500000.0"],
                            "times": ["0.0", "1E-7", "0.0000039", "0.000004"],
                        },
                        "pattern": "uniform",
                    },
                }
            ],
            "localDetuning": [],
        },
    }



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining GateModelProgram Class

"""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Optional

import numpy as np

from qbraid.programs.program import QuantumProgram

if TYPE_CHECKING:
    import qbraid.runtime


class GateModelProgram(QuantumProgram, ABC):
    """Abstract class for qbraid program wrapper objects."""

    @property
    @abstractmethod
    def qubits(self) -> list[Any]:
        """Return the qubits acted upon by the operations in this circuit"""

    @property
    def num_qubits(self) -> int:
        """Return the number of qubits in the circuit."""
        return len(self.qubits)

    @property
    @abstractmethod
    def num_clbits(self) -> Optional[int]:
        """Return the number of classical bits in the circuit."""

    @property
    def depth(self) -> int:
        """Return the circuit depth (i.e., length of critical path)."""
        raise NotImplementedError

    def _unitary(self) -> np.ndarray:
        """Calculate unitary of circuit."""
        raise NotImplementedError

    def unitary(self) -> np.ndarray:
        """Calculate unitary of circuit."""
        if self.spec.alias in ["pyquil", "qiskit", "qasm3"]:
            return self.unitary_rev_qubits()
        return self._unitary()

    def unitary_rev_qubits(self) -> np.ndarray:
        """Performs Kronecker (tensor) product factor permutation of given matrix.
        Returns a matrix equivalent to that computed from a quantum circuit if its
        qubit indicies were reversed.

        Args:
            matrix (np.ndarray): The input matrix, assumed to be a 2^N x 2^N square matrix
                                where N is an integer.

        Returns:
            np.ndarray: The matrix with permuted Kronecker product factors.

        Raises:
            ValueError: If the input matrix is not square or its size is not a power of 2.
        """
        matrix = self._unitary()
        if matrix.shape[0] != matrix.shape[1] or (matrix.shape[0] & (matrix.shape[0] - 1)) != 0:
            raise ValueError("Input matrix must be a square matrix of size 2^N for some integer N.")

        # Determine the number of qubits from the matrix size
        num_qubits = int(np.log2(matrix.shape[0]))

        # Create an empty matrix of the same size
        permuted_matrix = np.zeros((2**num_qubits, 2**num_qubits), dtype=complex)

        for i in range(2**num_qubits):
            for j in range(2**num_qubits):
                # pylint: disable=consider-using-generator
                # Convert indices to binary representations (qubit states)
                bits_i = [((i >> bit) & 1) for bit in range(num_qubits)]
                bits_j = [((j >> bit) & 1) for bit in range(num_qubits)]

                # Reverse the bits
                reversed_i = sum([bit << (num_qubits - 1 - k) for k, bit in enumerate(bits_i)])
                reversed_j = sum([bit << (num_qubits - 1 - k) for k, bit in enumerate(bits_j)])

                # Update the new matrix
                permuted_matrix[reversed_i, reversed_j] = matrix[i, j]

        return permuted_matrix

    def unitary_little_endian(self) -> np.ndarray:
        """Converts unitary calculated using big-endian system to its
        equivalent form in a little-endian system.

        Args:
            matrix: big-endian unitary

        Raises:
            ValueError: If input matrix is not unitary

        Returns:
            little-endian unitary

        """
        matrix = self.unitary()
        rank = len(matrix)
        if not np.allclose(np.eye(rank), matrix.dot(matrix.T.conj())):
            raise ValueError("Input matrix must be unitary.")
        num_qubits = int(np.log2(rank))
        tensor_be = matrix.reshape([2] * 2 * num_qubits)
        indicies_in = list(reversed(range(num_qubits)))
        indicies_out = [i + num_qubits for i in indicies_in]
        tensor_le = np.einsum(tensor_be, indicies_in + indicies_out)
        return tensor_le.reshape([rank, rank])

    def remove_idle_qubits(self) -> None:
        """Remove empty registers of circuit."""
        raise NotImplementedError

    def reverse_qubit_order(self) -> None:
        """Rerverse qubit ordering of circuit."""
        raise NotImplementedError

    def transform(self, device: qbraid.runtime.QuantumDevice) -> None:
        """Transform program to according to device target profile."""
        return None

    def serialize(self) -> dict[str, str]:
        """Return the program in a format suitable for submission to the qBraid API."""
        raise NotImplementedError



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for conversions between Cirq circuits and pytket circuits.

"""
import pytest

try:
    import numpy as np
    from cirq import Circuit, LineQubit, ops, testing
    from pytket.circuit import Circuit as TKCircuit
    from pytket.qasm import circuit_to_qasm_str

    from qbraid.interface import circuits_allclose
    from qbraid.transpiler.conversions.qasm2 import qasm2_to_cirq
    from qbraid.transpiler.converter import transpile

    from ..cirq_utils import _equal

    pytket_not_installed = False
except ImportError:
    pytket_not_installed = True

pytestmark = pytest.mark.skipif(pytket_not_installed, reason="pytket not installed")


def test_bell_state_to_from_circuits():
    """Tests cirq.Circuit --> pytket.circuit.Circuit --> cirq.Circuit
    with a Bell state circuit.
    """
    qreg = LineQubit.range(2)
    cirq_circuit = Circuit([ops.H.on(qreg[0]), ops.CNOT.on(qreg[0], qreg[1])])
    pytket_circuit = transpile(cirq_circuit, "pytket")  # pytket from Cirq
    circuit_cirq = transpile(pytket_circuit, "cirq")  # Cirq from pytket
    assert np.allclose(cirq_circuit.unitary(), circuit_cirq.unitary())


def test_random_circuit_to_from_circuits():
    """Tests cirq.Circuit --> pytket.circuit.Circuit --> cirq.Circuit
    with a random two-qubit circuit.
    """
    cirq_circuit = testing.random_circuit(qubits=2, n_moments=10, op_density=0.99, random_state=1)
    pytket_circuit = transpile(cirq_circuit, "pytket")
    circuit_cirq = transpile(pytket_circuit, "cirq")
    assert np.allclose(cirq_circuit.unitary(), circuit_cirq.unitary())


@pytest.mark.parametrize("as_qasm", (True, False))
def test_convert_with_barrier(as_qasm):
    """Tests converting a pytket circuit with a barrier to a Cirq circuit."""
    n = 5
    pytket_circuit = TKCircuit(n)
    pytket_circuit.add_barrier(list(range(n)))

    if as_qasm:
        cirq_circuit = qasm2_to_cirq(circuit_to_qasm_str(pytket_circuit))
    else:
        cirq_circuit = transpile(pytket_circuit, "cirq")

    assert _equal(cirq_circuit, Circuit())


@pytest.mark.parametrize("as_qasm", (True, False))
def test_convert_with_multiple_barriers(as_qasm):
    """Tests converting a pytket circuit with barriers to a Cirq circuit."""
    n = 1
    num_ops = 10

    pytket_circuit = TKCircuit(n)
    for _ in range(num_ops):
        for i in range(n):
            pytket_circuit.H(i)
        pytket_circuit.add_barrier(list(range(n)))

    if as_qasm:
        cirq_circuit = qasm2_to_cirq(circuit_to_qasm_str(pytket_circuit))
    else:
        cirq_circuit = transpile(pytket_circuit, "cirq")

    qbit = LineQubit(0)
    correct = Circuit(ops.H.on(qbit) for _ in range(num_ops))
    assert _equal(cirq_circuit, correct)
    assert circuits_allclose(pytket_circuit, cirq_circuit, strict_gphase=True)



# Copyright (C) 2025 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name
"""
Unit tests for Azure Quantum runtime (remote)

"""
from __future__ import annotations

import importlib.util
import os
from typing import TYPE_CHECKING
from unittest.mock import MagicMock, patch

import pytest
from azure.quantum import Workspace

from qbraid.runtime import (
    AhsResultData,
    AzureQuantumProvider,
    DeviceStatus,
    GateModelResultData,
    JobStatus,
    Result,
)

# Skip pyquil and/or pulser tests if not installed
pyquil_found = importlib.util.find_spec("pyquil") is not None
pulser_found = importlib.util.find_spec("pulser") is not None

if TYPE_CHECKING:
    import pulser as pulser_
    import pyquil as pyquil_


@pytest.mark.remote
@pytest.mark.skip(reason="The Free Credits plan has been discontinued in Azure Quantum")
def test_submit_qasm2_to_quantinuum(provider: AzureQuantumProvider):
    """Test submitting an OpenQASM 2 string to run on the Quantinuum simulator."""
    device = provider.get_device("quantinuum.sim.h1-1sc")
    status = device.status()

    if status != DeviceStatus.ONLINE:
        pytest.skip(f"{device.id} is {status.value}")

    circuit = """
    OPENQASM 2.0;
    include "qelib1.inc";
    qreg q[3];
    creg c0[3];
    h q[0];
    cx q[0], q[1];
    cx q[1], q[2];
    measure q[0] -> c0[0];
    measure q[1] -> c0[1];
    measure q[2] -> c0[2];
    """

    job = device.run(circuit, shots=100)
    job.wait_for_final_state()
    assert job.status() == JobStatus.COMPLETED

    result = job.result()
    assert isinstance(result, Result)
    assert isinstance(result.data, GateModelResultData)
    assert result.data.get_counts() == {"000": 100}


@pytest.mark.remote
@pytest.mark.skip(reason="The Free Credits plan has been discontinued in Azure Quantum")
def test_submit_json_to_ionq(provider: AzureQuantumProvider):
    """Test submitting a circuit JSON to run on the IonQ simulator."""
    device = provider.get_device("ionq.simulator")
    status = device.status()

    if status != DeviceStatus.ONLINE:
        pytest.skip(f"{device.id} is {status.value}")

    circuit = {
        "qubits": 3,
        "circuit": [
            {"gate": "h", "target": 0},
            {"gate": "cnot", "control": 0, "target": 1},
            {"gate": "cnot", "control": 0, "target": 2},
        ],
    }

    job = device.run(circuit, shots=100)
    job.wait_for_final_state()
    assert job.status() == JobStatus.COMPLETED

    result = job.result()
    assert isinstance(result, Result)
    assert isinstance(result.data, GateModelResultData)
    assert result.data.get_counts() == {"000": 50, "111": 50}


@pytest.fixture
@pytest.mark.skipif(not pyquil_found, reason="pyquil not installed")
def pyquil_program() -> pyquil_.Program:
    """Fixture for a PyQuil program."""
    import pyquil

    p = pyquil.Program()
    ro = p.declare("ro", "BIT", 2)
    p += pyquil.gates.H(0)
    p += pyquil.gates.CNOT(0, 1)
    p += pyquil.gates.MEASURE(0, ro[0])
    p += pyquil.gates.MEASURE(1, ro[1])

    return p


@pytest.fixture
@pytest.mark.skipif(not pulser_found, reason="pulser not installed")
def pulser_sequence() -> pulser_.Sequence:
    """Fixture for a Pulser sequence."""
    # pylint: disable=import-outside-toplevel
    import numpy as np
    import pulser
    from pulser.waveforms import BlackmanWaveform, RampWaveform

    # pylint: enable=import-outside-toplevel

    qubits = {
        "q0": (0, 0),
        "q1": (0, 10),
        "q2": (8, 2),
        "q3": (1, 15),
        "q4": (-10, -3),
        "q5": (-8, 5),
    }
    reg = pulser.Register(qubits)

    seq = pulser.Sequence(reg, pulser.devices.DigitalAnalogDevice)
    # Declare a channel. In this example we will be using `rydberg_global`
    seq.declare_channel("ch0", "rydberg_global")
    amp_wf = BlackmanWaveform(1000, np.pi)
    det_wf = RampWaveform(1000, -5, 5)
    pulse = pulser.Pulse(amp_wf, det_wf, 0)
    seq.add(pulse, "ch0")

    return seq


@pytest.fixture
@pytest.mark.skipif(not pyquil_found, reason="pyquil not installed")
def quil_string(pyquil_program: pyquil_.Program) -> str:
    """Fixture for a Quil string."""
    return pyquil_program.out()


@pytest.mark.remote
@pytest.mark.skipif(not pyquil_found, reason="pyquil not installed")
@pytest.mark.parametrize("direct", [(True), (False)])
@pytest.mark.skip(reason="The Free Credits plan has been discontinued in Azure Quantum")
def test_submit_quil_to_rigetti(
    provider: AzureQuantumProvider, pyquil_program: pyquil_.Program, quil_string: str, direct: bool
):
    """Test submitting a pyQuil program or Quil string to run on the Rigetti simulator."""
    device = provider.get_device("rigetti.sim.qvm")
    status = device.status()

    if status != DeviceStatus.ONLINE:
        pytest.skip(f"{device.id} is {status.value}")

    shots = 100
    input_params = {}

    if direct:
        job = device.submit(quil_string, shots=shots, input_params=input_params)
    else:
        job = device.run(pyquil_program, shots=100, input_params=input_params)

    job.wait_for_final_state()
    assert job.status() == JobStatus.COMPLETED

    result = job.result()
    assert isinstance(result, Result)
    assert isinstance(result.data, GateModelResultData)
    assert result.data.get_counts() == {"00": 60, "11": 40}


@pytest.mark.remote
@pytest.mark.skipif(not pulser_found, reason="pulser not installed")
def test_submit_sequence_to_pasqal(
    provider: AzureQuantumProvider, pulser_sequence: pulser_.Sequence
):
    """Test submitting a Pulser sequence-builder to run on the Pulser emulator."""
    try:
        device = provider.get_device("pasqal.sim.emu-tn")
    except ValueError:
        pytest.skip("Pasqal device not available in the current workspace")

    status = device.status()

    if status != DeviceStatus.ONLINE:
        pytest.skip(f"{device.id} is {status.value}")

    shots = 100
    input_params = {}

    job = device.submit(pulser_sequence, shots=shots, input_params=input_params)

    job.wait_for_final_state()
    assert job.status() == JobStatus.COMPLETED

    result = job.result()
    assert isinstance(result, Result)
    assert isinstance(result.data, AhsResultData)
    assert result.data.get_counts() == {"00": 60, "11": 40}


@pytest.mark.remote
@pytest.mark.skipif(
    not os.getenv("AZURE_QUANTUM_CONNECTION_STRING"), reason="No connection string set"
)
def test_provider_from_env_variables():
    """Test creating a provider from environment variables."""
    provider = AzureQuantumProvider()
    assert isinstance(provider.workspace, Workspace)


def test_workspace_from_connection_string():
    """Test creating a workspace from a connection string."""
    mock_connection_string = "mock_connection_string"

    with (
        patch.dict(os.environ, {"AZURE_QUANTUM_CONNECTION_STRING": mock_connection_string}),
        patch("azure.quantum.Workspace.from_connection_string") as mock_from_connection_string,
    ):

        mock_workspace = MagicMock(spec=Workspace)
        mock_from_connection_string.return_value = mock_workspace

        provider = AzureQuantumProvider()

        mock_from_connection_string.assert_called_once_with(mock_connection_string)
        assert provider.workspace == mock_workspace



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for transforming and extracting data from OpenQASM programs.

.. currentmodule:: qbraid.passes.qasm

Functions
----------

.. autosummary::
   :toctree: ../stubs/

    rebase
    insert_gate_def
    replace_gate_names
    add_stdgates_include
    remove_stdgates_include
    convert_qasm_pi_to_decimal
    normalize_qasm_gate_params

"""
from .compat import (
    add_stdgates_include,
    convert_qasm_pi_to_decimal,
    insert_gate_def,
    normalize_qasm_gate_params,
    remove_stdgates_include,
    replace_gate_names,
)
from .decompose import rebase

__all__ = [
    "rebase",
    "insert_gate_def",
    "replace_gate_names",
    "add_stdgates_include",
    "remove_stdgates_include",
    "convert_qasm_pi_to_decimal",
    "normalize_qasm_gate_params",
]



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining exceptions for errors raised during conversions

"""
from typing import Optional

from qbraid.exceptions import QbraidError


class ProgramConversionError(QbraidError):
    """Base class for errors raised while converting a circuit."""


class NodeNotFoundError(ValueError, QbraidError):
    """Class for errors raised when a node is not present in a ConversionGraph."""

    def __init__(self, graph_type: str, package: str, nodes: list[str]):
        message = (
            f"{graph_type} conversion graph does not contain node '{package}'. "
            f"Supported nodes are: {nodes}"
        )
        super().__init__(message)


class ConversionPathNotFoundError(QbraidError):
    """Class for errors raised when there is no path between two nodes in a ConversionGraph."""

    def __init__(self, source: str, target: str, max_depth: Optional[int] = None):
        max_depth_msg = f" with depth <= {max_depth}" if max_depth is not None else ""
        message = f"No conversion path found from '{source}' to '{target}'{max_depth_msg}"
        super().__init__(message)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining abstract QuantumJob Class

"""
from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from time import sleep, time
from typing import TYPE_CHECKING, Any, Optional

from .enums import JobStatus
from .exceptions import ResourceNotFoundError

if TYPE_CHECKING:
    import qbraid.runtime
    from qbraid.runtime.result_data import ResultDataType


class QuantumJob(ABC):
    """Abstract interface for job-like classes."""

    def __init__(
        self, job_id: str | int, device: Optional[qbraid.runtime.QuantumDevice] = None, **kwargs
    ):
        self._job_id = job_id
        self._device = device
        self._cache_metadata = {"job_id": job_id, **kwargs}

    @property
    def id(self) -> str | int:  # pylint: disable=invalid-name
        """Return a unique id identifying the job."""
        return self._job_id

    @property
    def device(self) -> qbraid.runtime.QuantumDevice:
        """Returns the qbraid QuantumDevice object associated with this job."""
        if self._device is None:
            raise ResourceNotFoundError("Job does not have an associated device.")
        return self._device

    def is_terminal_state(self) -> bool:
        """Returns True if job is in final state. False otherwise."""
        terminal_states = JobStatus.terminal_states()
        if self._cache_metadata.get("status", None) in terminal_states:
            return True

        status = self.status()
        return status in terminal_states

    @abstractmethod
    def status(self) -> JobStatus:
        """Return the status of the job / task , among the values of ``JobStatus``."""

    def metadata(self) -> dict[str, Any]:
        """Return the metadata regarding the job."""
        status = self.status()
        self._cache_metadata["status"] = status
        return self._cache_metadata

    def wait_for_final_state(self, timeout: Optional[int] = None, poll_interval: int = 5) -> None:
        """Poll the job status until it progresses to a final state.

        Args:
            timeout: Seconds to wait for the job. If ``None``, wait indefinitely.
            poll_interval: Seconds between queries. Defaults to 5 seconds.

        Raises:
            JobStateError: If the job does not reach a final state before the specified timeout.

        """
        start_time = time()
        while not self.is_terminal_state():
            elapsed_time = time() - start_time
            if timeout is not None and elapsed_time >= timeout:
                raise TimeoutError(f"Timeout while waiting for job {self.id}.")
            sleep(poll_interval)

    async def _wait_for_final_state(
        self, timeout: Optional[int] = None, poll_interval: int = 5
    ) -> None:
        """Asynchronously wait for the job to reach a terminal state (e.g., COMPLETED, FAILED).

        This non-blocking method uses asyncio to periodically poll the job's status,
        allowing other coroutines to run while waiting. It is especially useful in
        asynchronous applications where blocking the event loop is undesirable.

        Args:
            timeout (Optional[int]): Maximum number of seconds to wait for the job.
                If None, waits indefinitely.
            poll_interval (int): Seconds between queries. Defaults to 5.

        Raises:
            TimeoutError: If the job does not reach a terminal state before the specified timeout.
        """
        start_time = time()
        while not self.is_terminal_state():
            elapsed_time = time() - start_time
            if timeout is not None and elapsed_time >= timeout:
                raise TimeoutError(f"Timeout while waiting for job {self.id}.")
            await asyncio.sleep(poll_interval)

    async def async_result(
        self, timeout: Optional[int] = None, poll_interval: int = 5
    ) -> qbraid.runtime.Result[ResultDataType]:
        """Asynchronously wait for the job to reach a final state and return the result.

        This method provides a non-blocking way to poll the job status using asyncio.
        It allows developers to `await` the job without blocking the event loop,
        making it suitable for integration in async applications and pipelines.

        Args:
            timeout (Optional[int]): Maximum number of seconds to wait for the job.
                If None, waits indefinitely.
            poll_interval (int): Seconds between status checks. Defaults to 5.

        Returns:
            Result[ResultDataType]: The result object associated with the job,
            if successfully completed.

        Raises:
            TimeoutError: If the job does not reach a terminal state before the timeout expires.
        """
        await self._wait_for_final_state(timeout, poll_interval)
        return self.result()

    @abstractmethod
    def result(self) -> qbraid.runtime.Result[ResultDataType]:
        """Return the results of the job."""

    @abstractmethod
    def cancel(self) -> None:
        """Attempt to cancel the job."""

    def __repr__(self) -> str:
        """String representation of a QuantumJob object."""
        return f"<{self.__class__.__name__}(id:'{self.id}')>"



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for the qbraid transpiler conversions module.

"""
from typing import Optional

import cirq
import numpy as np
import pytest

from qbraid.programs import NATIVE_REGISTRY, load_program
from qbraid.transpiler.conversions import conversion_functions
from qbraid.transpiler.converter import transpile
from qbraid.transpiler.graph import ConversionGraph


def find_cirq_targets(skip: Optional[list[str]] = None):
    """Find all Cirq conversion targets."""
    skip = skip or []
    cirq_targets = []
    for function in conversion_functions:
        if function.startswith("cirq_to_"):
            _, target_library = function.split("_to_")
            if target_library not in skip and target_library in NATIVE_REGISTRY:
                cirq_targets.append(target_library)
    return cirq_targets


TARGETS = find_cirq_targets()


@pytest.mark.parametrize("frontend", TARGETS)
def test_convert_circuit_operation_from_cirq(frontend):
    """Test converting Cirq FrozenCircuit operation to OpenQASM"""
    q = cirq.NamedQubit("q")
    cirq_circuit = cirq.Circuit(
        cirq.Y(q), cirq.CircuitOperation(cirq.FrozenCircuit(cirq.X(q)), repetitions=5), cirq.Z(q)
    )

    graph = ConversionGraph()

    if not graph.has_path("cirq", frontend):
        pytest.skip(f"conversion from cirq to {frontend} not yet supported")

    test_circuit = transpile(cirq_circuit, frontend, conversion_graph=graph)

    cirq_unitary = load_program(cirq_circuit).unitary()

    try:
        test_unitary = load_program(test_circuit).unitary()
    except NotImplementedError:
        pytest.skip(f"Unitary calculation not implemented for {frontend}")

    assert np.allclose(cirq_unitary, test_unitary)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Qiskit conversions

.. currentmodule:: qbraid.transpiler.conversions.qiskit

Functions
----------

.. autosummary::
   :toctree: ../stubs/

   qiskit_to_qasm2
   qiskit_to_qasm3
   qiskit_to_braket
   qiskit_to_pyqir
   qiskit_to_ionq

"""
from .qiskit_extras import qiskit_to_braket, qiskit_to_ionq, qiskit_to_pyqir
from .qiskit_to_qasm2 import qiskit_to_qasm2
from .qiskit_to_qasm3 import qiskit_to_qasm3

__all__ = [
    "qiskit_to_qasm2",
    "qiskit_to_qasm3",
    "qiskit_to_braket",
    "qiskit_to_pyqir",
    "qiskit_to_ionq",
]



# Copyright (C) 2024 qBraid
# Copyright (C) 2020 The Cirq Developers
#
# This file is part of the qBraid-SDK.
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. This specific file, adapted from Cirq, is dual-licensed under both the
# Apache License, Version 2.0, and the GPL v3. You may not use this file except in
# compliance with the applicable license. You may obtain a copy of the Apache License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# This file includes code adapted from Cirq (https://github.com/quantumlib/Cirq)
# with modifications by qBraid. The original copyright notice is included above.
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# qbraid: skip-header

"""
Module for testing qBraid quil input.

"""
import numpy as np
import pytest
from cirq import Circuit, LineQubit
from cirq.ops import (
    CCNOT,
    CNOT,
    CSWAP,
    CZ,
    ISWAP,
    SWAP,
    H,
    I,
    MeasurementGate,
    S,
    T,
    X,
    Y,
    Z,
    rx,
    ry,
    rz,
)

try:
    from pyquil import Program
    from pyquil.simulation.tools import program_unitary

    from qbraid.transpiler.conversions.pyquil.cirq_quil_input import (
        UndefinedQuilGate,
        UnsupportedQuilInstruction,
        circuit_from_quil,
        cphase,
        cphase00,
        cphase01,
        cphase10,
        pswap,
        xy,
    )

    pyquil_not_installed = False
except ImportError:
    pyquil_not_installed = True

pytestmark = pytest.mark.skipif(pyquil_not_installed, reason="pyquil not installed")

QUIL_PROGRAM = """
DECLARE ro BIT[3]
I 0
I 1
I 2
X 0
Y 1
Z 2
H 0
S 1
T 2
PHASE(pi/8) 0
PHASE(pi/8) 1
PHASE(pi/8) 2
RX(pi/2) 0
RY(pi/2) 1
RZ(pi/2) 2
CZ 0 1
CNOT 1 2
CPHASE(pi/2) 0 1
CPHASE00(pi/2) 1 2
CPHASE01(pi/2) 0 1
CPHASE10(pi/2) 1 2
ISWAP 0 1
PSWAP(pi/2) 1 2
SWAP 0 1
XY(pi/2) 1 2
CCNOT 0 1 2
CSWAP 0 1 2
MEASURE 0 ro[0]
MEASURE 1 ro[1]
MEASURE 2 ro[2]
"""


def test_circuit_from_quil():
    """Test conversion of a Quil program to a Cirq Circuit."""
    q0, q1, q2 = LineQubit.range(3)
    cirq_circuit = Circuit(
        [
            I(q0),
            I(q1),
            I(q2),
            X(q0),
            Y(q1),
            Z(q2),
            H(q0),
            S(q1),
            T(q2),
            Z(q0) ** (1 / 8),
            Z(q1) ** (1 / 8),
            Z(q2) ** (1 / 8),
            rx(np.pi / 2)(q0),
            ry(np.pi / 2)(q1),
            rz(np.pi / 2)(q2),
            CZ(q0, q1),
            CNOT(q1, q2),
            cphase(np.pi / 2)(q0, q1),
            cphase00(np.pi / 2)(q1, q2),
            cphase01(np.pi / 2)(q0, q1),
            cphase10(np.pi / 2)(q1, q2),
            ISWAP(q0, q1),
            pswap(np.pi / 2)(q1, q2),
            SWAP(q0, q1),
            xy(np.pi / 2)(q1, q2),
            CCNOT(q0, q1, q2),
            CSWAP(q0, q1, q2),
            MeasurementGate(1, key="ro[0]")(q0),
            MeasurementGate(1, key="ro[1]")(q1),
            MeasurementGate(1, key="ro[2]")(q2),
        ]
    )
    # build the same Circuit, using Quil
    quil_circuit = circuit_from_quil(QUIL_PROGRAM)
    # test Circuit equivalence
    assert cirq_circuit == quil_circuit

    pyquil_circuit = Program(QUIL_PROGRAM)
    # drop declare and measures, get Program unitary
    pyquil_unitary = program_unitary(pyquil_circuit[1:-3], n_qubits=3)
    # fix qubit order convention
    cirq_circuit_swapped = Circuit(SWAP(q0, q2), cirq_circuit[:-1], SWAP(q0, q2))
    # get Circuit unitary
    cirq_unitary = cirq_circuit_swapped.unitary()
    # test unitary equivalence
    assert np.isclose(pyquil_unitary, cirq_unitary).all()


QUIL_PROGRAM_WITH_DEFGATE = """
DEFGATE MYZ:
    1,0
    0,-1

X 0
MYZ 0
"""


def test_quil_with_defgate():
    """Test conversion of a Quil program with DEFGATE to a Cirq Circuit."""
    q0 = LineQubit(0)
    cirq_circuit = Circuit([X(q0), Z(q0)])
    quil_circuit = circuit_from_quil(QUIL_PROGRAM_WITH_DEFGATE)
    assert np.isclose(quil_circuit.unitary(), cirq_circuit.unitary()).all()


QUIL_PROGRAM_WITH_PARAMETERIZED_DEFGATE = """
DEFGATE MYPHASE(%phi):
    1,0
    0,EXP(i*%phi)

X 0
MYPHASE 0
"""


def test_unsupported_quil_instruction():
    """Test that unsupported Quil instructions raise an exception."""

    with pytest.raises(UnsupportedQuilInstruction):
        circuit_from_quil("NOP")

    with pytest.raises(UnsupportedQuilInstruction):
        circuit_from_quil('PRAGMA ADD-KRAUS X 0 "(0.0 1.0 1.0 0.0)"')

    with pytest.raises(UnsupportedQuilInstruction):
        circuit_from_quil("RESET")

    with pytest.raises(UnsupportedQuilInstruction):
        circuit_from_quil(QUIL_PROGRAM_WITH_PARAMETERIZED_DEFGATE)


def test_undefined_quil_gate():
    """There are no such things as FREDKIN & TOFFOLI in Quil. The standard
    names for those gates in Quil are CSWAP and CCNOT. Of course, they can
    be defined via DEFGATE / DEFCIRCUIT."""
    with pytest.raises(UndefinedQuilGate):
        circuit_from_quil("FREDKIN 0 1 2")

    with pytest.raises(UndefinedQuilGate):
        circuit_from_quil("TOFFOLI 0 1 2")


def test_measurement_without_classical_reg():
    """Measure operations must declare a classical register."""
    with pytest.raises(UnsupportedQuilInstruction):
        circuit_from_quil("MEASURE 0")



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name,unused-argument

"""
Unit tests for submitting a Braket AHS job.

"""
import json
import uuid
from typing import Any, Optional, Union
from unittest.mock import MagicMock, Mock, patch

import numpy as np
import pytest
from braket.ahs.analog_hamiltonian_simulation import AnalogHamiltonianSimulation
from braket.ahs.atom_arrangement import AtomArrangement
from braket.ahs.driving_field import DrivingField
from braket.aws import AwsQuantumTask
from braket.aws.aws_quantum_task_batch import AwsQuantumTaskBatch
from braket.device_schema.quera.quera_device_capabilities_v1 import QueraDeviceCapabilities
from braket.devices import Devices
from braket.ir.ahs import Program
from braket.task_result import (
    AdditionalMetadata,
    AnalogHamiltonianSimulationShotMeasurement,
    AnalogHamiltonianSimulationShotMetadata,
    AnalogHamiltonianSimulationShotResult,
    AnalogHamiltonianSimulationTaskResult,
    TaskMetadata,
)
from braket.tasks import AnalogHamiltonianSimulationQuantumTaskResult
from braket.timings.time_series import TimeSeries

from qbraid.programs import ExperimentType, ProgramSpec
from qbraid.runtime import DeviceStatus, TargetProfile
from qbraid.runtime.aws.device import BraketDevice
from qbraid.runtime.aws.job import BraketQuantumTask
from qbraid.runtime.aws.provider import BraketProvider
from qbraid.runtime.aws.result_builder import BraketAhsResultBuilder, ResultDecodingError
from qbraid.runtime.exceptions import DeviceProgramTypeMismatchError

AQUILA_ARN = Devices.QuEra.Aquila
TASK_ARN = "arn:aws:braket:us-east-1:0123456789012:quantum-task/" + str(uuid.uuid4())

PARADIGM = {
    "braketSchemaHeader": {
        "name": "braket.device_schema.quera.quera_ahs_paradigm_properties",
        "version": "1",
    },
    "qubitCount": 256,
    "lattice": {
        "area": {"width": 7.5e-05, "height": 7.6e-05},
        "geometry": {
            "spacingRadialMin": 4e-06,
            "spacingVerticalMin": 4e-06,
            "positionResolution": 1e-08,
            "numberSitesMax": 256,
        },
    },
    "rydberg": {
        "c6Coefficient": 5.42e-24,
        "rydbergGlobal": {
            "rabiFrequencyRange": [0.0, 15800000.0],
            "rabiFrequencyResolution": 400.0,
            "rabiFrequencySlewRateMax": 400000000000000.0,
            "detuningRange": [-125000000.0, 125000000.0],
            "detuningResolution": 0.2,
            "detuningSlewRateMax": 6000000000000000.0,
            "phaseRange": [-99.0, 99.0],
            "phaseResolution": 5e-07,
            "timeResolution": 1e-09,
            "timeDeltaMin": 5e-08,
            "timeMin": 0.0,
            "timeMax": 4e-06,
        },
    },
    "performance": {
        "lattice": {
            "positionErrorAbs": 2.25e-07,
            "sitePositionError": 1e-07,
            "atomPositionError": 2e-07,
            "fillingErrorTypical": 0.008,
            "fillingErrorWorst": 0.05,
            "vacancyErrorTypical": 0.001,
            "vacancyErrorWorst": 0.005,
            "atomLossProbabilityTypical": 0.005,
            "atomLossProbabilityWorst": 0.01,
            "atomCaptureProbabilityTypical": 0.001,
            "atomCaptureProbabilityWorst": 0.002,
            "atomDetectionErrorFalsePositiveTypical": 0.001,
            "atomDetectionErrorFalsePositiveWorst": 0.005,
            "atomDetectionErrorFalseNegativeTypical": 0.001,
            "atomDetectionErrorFalseNegativeWorst": 0.005,
        },
        "rydberg": {
            "rydbergGlobal": {
                "rabiFrequencyErrorRel": 0.03,
                "rabiFrequencyGlobalErrorRel": 0.02,
                "rabiFrequencyInhomogeneityRel": 0.02,
                "groundDetectionError": 0.05,
                "rydbergDetectionError": 0.1,
                "groundPrepError": 0.01,
                "rydbergPrepErrorBest": 0.05,
                "rydbergPrepErrorWorst": 0.07,
                "T1Single": 7.5e-05,
                "T1Ensemble": 7.5e-05,
                "T2StarSingle": 5e-06,
                "T2StarEnsemble": 4.75e-06,
                "T2EchoSingle": 8e-06,
                "T2EchoEnsemble": 7e-06,
                "T2RabiSingle": 8e-06,
                "T2RabiEnsemble": 7e-06,
                "T2BlockadedRabiSingle": 8e-06,
                "T2BlockadedRabiEnsemble": 7e-06,
                "detuningError": 1000000.0,
                "detuningInhomogeneity": 1000000.0,
                "rabiAmplitudeRampCorrection": [
                    {"rampTime": 5e-08, "rabiCorrection": 0.92},
                    {"rampTime": 7.5e-08, "rabiCorrection": 0.97},
                    {"rampTime": 1e-07, "rabiCorrection": 1.0},
                ],
            }
        },
    },
}

CAPABILITIES = {
    "service": {
        "braketSchemaHeader": {
            "name": "braket.device_schema.device_service_properties",
            "version": "1",
        },
        "executionWindows": [
            {"executionDay": "Monday", "windowStartHour": "01:00:00", "windowEndHour": "23:59:59"},
            {"executionDay": "Tuesday", "windowStartHour": "00:00:00", "windowEndHour": "12:00:00"},
            {
                "executionDay": "Wednesday",
                "windowStartHour": "00:00:00",
                "windowEndHour": "12:00:00",
            },
            {"executionDay": "Friday", "windowStartHour": "00:00:00", "windowEndHour": "23:59:59"},
            {
                "executionDay": "Saturday",
                "windowStartHour": "00:00:00",
                "windowEndHour": "23:59:59",
            },
            {"executionDay": "Sunday", "windowStartHour": "00:00:00", "windowEndHour": "12:00:00"},
        ],
        "shotsRange": [1, 1000],
        "deviceCost": {"price": 0.01, "unit": "shot"},
        "deviceLocation": "Boston, USA",
        "updatedAt": "2024-06-03T12:00:00+00:00",
    },
    "action": {"braket.ir.ahs.program": {"version": ["1"], "actionType": "braket.ir.ahs.program"}},
    "deviceParameters": {},
    "braketSchemaHeader": {
        "name": "braket.device_schema.quera.quera_device_capabilities",
        "version": "1",
    },
    "paradigm": PARADIGM,
}

METADATA = {
    "deviceArn": "arn:aws:braket:us-east-1::device/qpu/quera/Aquila",
    "deviceCapabilities": json.dumps(CAPABILITIES),
    "deviceName": "Aquila",
    "deviceQueueInfo": [
        {"queue": "QUANTUM_TASKS_QUEUE", "queuePriority": "Normal", "queueSize": "20"},
        {"queue": "QUANTUM_TASKS_QUEUE", "queuePriority": "Priority", "queueSize": "0"},
        {"queue": "JOBS_QUEUE", "queueSize": "0"},
    ],
    "deviceStatus": "ONLINE",
    "deviceType": "QPU",
    "providerName": "QuEra",
}


class MockBotoSession:
    """Mock Boto session class."""

    @property
    def region_name(self):
        """Return the region."""
        return "us-east-1"


class MockAwsSession:
    """Test class for AWS session."""

    @property
    def region(self):
        """Return the region."""
        return "us-east-1"

    @property
    def boto_session(self):
        """Return the boto session."""
        return MockBotoSession()

    def get_device(self, arn):  # pylint: disable=unused-argument
        """Returns metadata for a device."""
        return METADATA

    def create_quantum_task(self, **kwargs):
        """Creates a quantum task."""
        return TASK_ARN

    def get_quantum_task(self, *args, **kwargs):
        """Returns a quantum task."""
        return AwsQuantumTask(arn=TASK_ARN, aws_session=self)


class MockAquilaDevice:
    """Test class for braket device."""

    @property
    def arn(self):
        """Return the ARN of the device."""
        return METADATA["deviceArn"]

    @property
    def name(self):
        """Return the name of the device."""
        return METADATA["deviceName"]

    @property
    def aws_session(self):
        """Return the AWS session."""
        return MockAwsSession()

    @property
    def status(self):
        """Return the status of the device."""
        return METADATA["deviceStatus"]

    @property
    def is_available(self):
        """Return whether the device is available."""
        return True

    @property
    def properties(self):
        """Return the properties of the device."""
        return QueraDeviceCapabilities(**CAPABILITIES)

    @staticmethod
    def get_device_region(arn):
        """Returns the region of a device."""
        return "us-east-1"

    def run_batch(
        self, task_specifications: Union[Any, list[Any]], shots: Optional[int] = None, **kwargs
    ):
        """Executes a batch of quantum tasks in parallel."""
        return AwsQuantumTaskBatch(
            aws_session=self.aws_session,
            device_arn=AQUILA_ARN,
            task_specifications=task_specifications,
            s3_destination_folder=("amazon-braket-qbraid-jobs", "tasks"),
            shots=shots,
            max_parallel=10,
            poll_timeout_seconds=AwsQuantumTask.DEFAULT_RESULTS_POLL_TIMEOUT,
            poll_interval_seconds=AwsQuantumTask.DEFAULT_RESULTS_POLL_INTERVAL,
            **kwargs,
        )


@pytest.fixture
def ahs_program():
    """Analogue Hamiltonian Simulation program."""
    separation = 5e-6
    block_separation = 15e-6
    k_max = 5
    m_max = 5

    register = AtomArrangement()
    for k in range(k_max):
        for m in range(m_max):
            register.add((block_separation * m, block_separation * k + separation / np.sqrt(3)))
            register.add(
                (
                    block_separation * m - separation / 2,
                    block_separation * k - separation / (2 * np.sqrt(3)),
                )
            )
            register.add(
                (
                    block_separation * m + separation / 2,
                    block_separation * k - separation / (2 * np.sqrt(3)),
                )
            )

    omega_const = 1.5e7  # rad/s
    rabi_pulse_area = np.pi / np.sqrt(3)  # rad
    omega_slew_rate_max = float(
        PARADIGM["rydberg"]["rydbergGlobal"]["rabiFrequencySlewRateMax"]
    )  # rad/s^2
    time_separation_min = float(PARADIGM["rydberg"]["rydbergGlobal"]["timeDeltaMin"])  # s

    amplitude = TimeSeries.trapezoidal_signal(
        rabi_pulse_area,
        omega_const,
        0.99 * omega_slew_rate_max,
        time_separation_min=time_separation_min,
    )

    detuning = TimeSeries.constant_like(amplitude, 0.0)
    phase = TimeSeries.constant_like(amplitude, 0.0)

    drive = DrivingField(amplitude=amplitude, detuning=detuning, phase=phase)
    ahs_program = AnalogHamiltonianSimulation(register=register, hamiltonian=drive)

    return ahs_program


@pytest.fixture
def ahs_result() -> AnalogHamiltonianSimulationQuantumTaskResult:
    """Mock Analog Hamiltonian Simulation task result."""
    task_metadata = TaskMetadata(**{"id": TASK_ARN, "deviceId": AQUILA_ARN, "shots": 100})
    additional_metadata = AdditionalMetadata(
        action=Program(
            setup={
                "ahs_register": {
                    "sites": [
                        [0.0, 0.0],
                        [0.0, 3.0e-6],
                        [0.0, 6.0e-6],
                        [3.0e-6, 0.0],
                        [3.0e-6, 3.0e-6],
                        [3.0e-6, 6.0e-6],
                    ],
                    "filling": [1, 1, 1, 1, 0, 0],
                }
            },
            hamiltonian={
                "drivingFields": [
                    {
                        "amplitude": {
                            "time_series": {
                                "values": [0.0, 2.51327e7, 2.51327e7, 0.0],
                                "times": [0.0, 3.0e-7, 2.7e-6, 3.0e-6],
                            },
                            "pattern": "uniform",
                        },
                        "phase": {
                            "time_series": {"values": [0, 0], "times": [0.0, 3.0e-6]},
                            "pattern": "uniform",
                        },
                        "detuning": {
                            "time_series": {
                                "values": [-1.25664e8, -1.25664e8, 1.25664e8, 1.25664e8],
                                "times": [0.0, 3.0e-7, 2.7e-6, 3.0e-6],
                            },
                            "pattern": "uniform",
                        },
                    }
                ],
                "localDetuning": [
                    {
                        "magnitude": {
                            "time_series": {
                                "values": [-1.25664e8, 1.25664e8],
                                "times": [0.0, 3.0e-6],
                            },
                            "pattern": [0.5, 1.0, 0.5, 0.5, 0.5, 0.5],
                        }
                    }
                ],
            },
        )
    )

    success_measurement = AnalogHamiltonianSimulationShotMeasurement(
        shotMetadata=AnalogHamiltonianSimulationShotMetadata(shotStatus="Success"),
        shotResult=AnalogHamiltonianSimulationShotResult(
            preSequence=[1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1],
            postSequence=[0, 0, 0, 1, 0, 1, 1, 0, 0, 1, 0],
        ),
    )

    success_measurement_extended = AnalogHamiltonianSimulationShotMeasurement(
        shotMetadata=AnalogHamiltonianSimulationShotMetadata(shotStatus="Success"),
        shotResult=AnalogHamiltonianSimulationShotResult(
            preSequence=[1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
            postSequence=[1, 0, 1, 1, 0, 1, 0, 0, 0, 0, 1],
        ),
    )

    measurements_extended = [success_measurement, success_measurement_extended]

    result = AnalogHamiltonianSimulationTaskResult(
        taskMetadata=task_metadata,
        additionalMetadata=additional_metadata,
        measurements=measurements_extended,
    )

    task_result = AnalogHamiltonianSimulationQuantumTaskResult.from_object(result)

    return task_result


@pytest.fixture
def aquila_profile():
    """Runtime profile for the device."""
    provider = BraketProvider()
    device = MockAquilaDevice()
    return provider._build_runtime_profile(device=device)


def test_aquila_runtime_profile(aquila_profile: TargetProfile):
    """Test building a runtime profile."""
    assert aquila_profile.device_id == METADATA["deviceArn"]
    assert aquila_profile.simulator is False if METADATA["deviceType"] == "QPU" else True
    assert aquila_profile.provider_name == METADATA["providerName"]


@pytest.mark.parametrize(
    "device_status, is_available, expected_status",
    [
        ("ONLINE", True, DeviceStatus.ONLINE),
        ("ONLINE", False, DeviceStatus.UNAVAILABLE),
        ("RETIRED", None, DeviceStatus.RETIRED),
        ("OFFLINE", None, DeviceStatus.OFFLINE),
    ],
)
def test_aquila_device_status(device_status, is_available, expected_status, aquila_profile):
    """Test getting Braket device status."""
    with patch("qbraid.runtime.aws.device.AwsDevice") as mock_aws_device:
        mock_aws_device_instance = mock_aws_device.return_value
        mock_aws_device_instance.status = device_status
        mock_aws_device_instance.is_available = is_available
        device = BraketDevice(profile=aquila_profile, session=MockAwsSession())
        assert device.status() == expected_status


@pytest.mark.parametrize(
    "quantum_tasks, expected_output",
    [
        ({"Normal": 5, "Priority": 3}, 8),
        ({"Normal": ">10", "Priority": "2"}, 12),
        ({"Normal": "4", "Priority": ">5"}, 9),
        ({"Normal": 0, "Priority": 0}, 0),
    ],
)
def test_queue_depth(quantum_tasks, expected_output, aquila_profile):
    """Test getting the number of jobs in the queue for the Braket device."""
    with patch("qbraid.runtime.aws.device.AwsDevice") as mock_aws_device:
        mock_aws_device_instance = mock_aws_device.return_value
        mock_queue_depth_info = MagicMock()
        mock_queue_depth_info.quantum_tasks = quantum_tasks
        mock_aws_device_instance.queue_depth.return_value = mock_queue_depth_info
        device = BraketDevice(profile=aquila_profile, session=MockAwsSession())
        assert device.queue_depth() == expected_output


def test_aquila_device_submit(aquila_profile, ahs_program):
    """Test getting a Braket device."""
    with patch("braket.aws.aws_quantum_task.AwsQuantumTask.state", return_value="COMPLETED"):
        device = BraketDevice(profile=aquila_profile, session=MockAwsSession())
        device._device = MockAquilaDevice()
        task = device.submit(ahs_program, shots=100)
        assert isinstance(task, BraketQuantumTask), f"Expected BraketQuantumTask, got {type(task)}"
        assert task.status().name == "COMPLETED"
        assert task.id == TASK_ARN


@patch("qbraid.runtime.aws.device.AwsDevice")
@patch("braket.ahs.analog_hamiltonian_simulation.AnalogHamiltonianSimulation.discretize")
def test_transform_ahs_programs(mock_aws_device, mock_discretize, aquila_profile, ahs_program):
    """Test transform method for device with AHS action type."""
    mock_aws_device.return_value = Mock()
    device = BraketDevice(aquila_profile)
    device.transform(ahs_program)
    mock_discretize.assert_called_once()


@patch("qbraid.runtime.aws.device.AwsDevice")
def test_transform_raises_for_mismatch(mock_aws_device, ahs_program):
    """Test raising exception for mismatched action type OPENQASM and program type AHS."""
    mock_aws_device.return_value = Mock()
    profile = TargetProfile(
        simulator=False,
        num_qubits=256,
        program_spec=ProgramSpec(AnalogHamiltonianSimulation, alias="braket_ahs"),
        provider_name="QuEra",
        device_id=AQUILA_ARN,
        experiment_type=ExperimentType.GATE_MODEL,
    )
    device = BraketDevice(profile)
    with pytest.raises(DeviceProgramTypeMismatchError):
        device.transform(ahs_program)


def test_get_counts_no_measurements():
    """Test getting counts with no measurements."""
    mock_result = MagicMock()
    mock_result.measurements = []
    job_result = BraketAhsResultBuilder(mock_result)
    counts = job_result.get_counts()
    assert counts is None


def test_get_counts(ahs_result):
    """Test getting counts from an AHS job result."""
    result = BraketAhsResultBuilder(ahs_result)
    counts = result.get_counts()
    expected_counts = {"rrrgeggrrgr": 1, "grggrgrrrrg": 1}
    assert counts == expected_counts


def test_get_counts_error():
    """Test getting counts from an AHS job result."""
    with patch(
        "qbraid.runtime.aws.result_builder.BraketAhsResultBuilder.measurements"
    ) as mock_measurements:
        mock_measurement = Mock()
        mock_measurement.status.name = "SUCCESS"
        del mock_measurement.pre_sequence
        mock_measurement.post_sequence = [0]

        mock_measurements.return_value = [mock_measurement]
        result = BraketAhsResultBuilder(Mock())

        with pytest.raises(ResultDecodingError):
            result.get_counts()



# Copyright (C) 2025 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining input and output data formats for Azure Quantum.

"""

from enum import Enum


class InputDataFormat(Enum):
    """Enum for defining input data formats for Azure Quantum."""

    MICROSOFT = "qir.v1"
    IONQ = "ionq.circuit.v1"
    QUANTINUUM = "honeywell.openqasm.v1"
    RIGETTI = "rigetti.quil.v1"
    PASQAL = "pasqal.pulser.v1"


class OutputDataFormat(Enum):
    """Enum for defining output data formats for Azure Quantum."""

    MICROSOFT_V1 = "microsoft.quantum-results.v1"
    MICROSOFT_V2 = "microsoft.quantum-results.v2"
    IONQ = "ionq.quantum-results.v1"
    QUANTINUUM = "honeywell.quantum-results.v1"
    RESOURCE_ESTIMATOR = "microsoft.resource-estimates.v1"
    RIGETTI = "rigetti.quil-results.v1"
    PASQAL = "pasqal.pulser-results.v1"



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name

"""
Unit tests for QuEra QASM simulator device.

"""
import importlib.util
from unittest.mock import PropertyMock, patch

import pytest
from pandas import DataFrame

from qbraid.programs import ExperimentType
from qbraid.runtime import ProgramValidationError, Result, TargetProfile
from qbraid.runtime.native import QbraidDevice, QbraidJob, QbraidProvider
from qbraid.runtime.native.result import QuEraQasmSimulatorResultData

from ._resources import JOB_DATA_QUERA_QASM, RESULTS_DATA_QUERA_QASM


@pytest.fixture
def mock_quera_profile():
    """Mock profile for testing."""
    return TargetProfile(
        device_id="quera_qasm_simulator",
        simulator=True,
        experiment_type=ExperimentType.GATE_MODEL,
        num_qubits=30,
        program_spec=QbraidProvider._get_program_spec("qasm2", "quera_qasm_simulator"),
    )


@pytest.fixture
def mock_quera_device(mock_quera_profile, mock_client):
    """Mock QuEra simulator device for testing."""
    return QbraidDevice(profile=mock_quera_profile, client=mock_client)


@pytest.mark.skipif(
    importlib.util.find_spec("flair_visual") is None, reason="flair-visual is not installed"
)
def test_quera_simulator_workflow(mock_provider, cirq_uniform, valid_qasm2_no_meas):
    """Test queara simulator job submission and result retrieval."""
    circuit = cirq_uniform(num_qubits=5, measure=False)
    num_qubits = len(circuit.all_qubits())

    provider = mock_provider
    device = provider.get_device("quera_qasm_simulator")

    shots = 10
    job = device.run(circuit, shots=shots, backend="cirq-gpu")
    assert isinstance(job, QbraidJob)
    assert job.is_terminal_state()
    assert job.queue_position() is None

    device._target_spec = None
    device.prepare = lambda x: {"openQasm": x}
    batch_job = device.run([valid_qasm2_no_meas], shots=shots)
    assert isinstance(batch_job, list)
    assert all(isinstance(job, QbraidJob) for job in batch_job)

    result = job.result()
    assert isinstance(result, Result)
    assert isinstance(result.data, QuEraQasmSimulatorResultData)
    assert repr(result.data).startswith("QuEraQasmSimulatorResultData")
    assert result.success
    assert result.job_id == JOB_DATA_QUERA_QASM["qbraidJobId"]
    assert result.device_id == JOB_DATA_QUERA_QASM["qbraidDeviceId"]
    assert result.data.backend == "cirq-gpu"

    counts = result.data.get_counts()
    probabilities = result.data.get_probabilities()
    assert len(counts) == len(probabilities) == 2
    assert sum(probabilities.values()) == 1.0
    assert result.data.measurements is None
    assert (
        result.data.flair_visual_version
        == RESULTS_DATA_QUERA_QASM["quera_simulation_result"]["flair_visual_version"]
    )
    assert result.data.backend == RESULTS_DATA_QUERA_QASM["backend"]

    logs = result.data.get_logs()
    assert isinstance(logs, DataFrame)

    assert result.details["shots"] == shots
    assert result.details["metadata"]["circuitNumQubits"] == num_qubits
    assert isinstance(result.details["timeStamps"]["executionDuration"], int)


def test_validate_quera_device_qasm_validator(mock_quera_device, valid_qasm2):
    """Test that validate method raises ValueError for QASM programs with measurements."""
    with pytest.raises(ProgramValidationError):
        mock_quera_device.validate([valid_qasm2])


def test_construct_aux_payload_no_spec(mock_quera_device, valid_qasm2_no_meas):
    """Test constructing auxiliary payload without a predefined program spec."""
    aux_payload = mock_quera_device._construct_aux_payload(valid_qasm2_no_meas)
    assert len(aux_payload) == 3
    assert aux_payload["openQasm"] == valid_qasm2_no_meas
    assert aux_payload["circuitNumQubits"] == 2
    assert aux_payload["circuitDepth"] == 3


def test_device_validate_emit_warning_for_num_qubits(mock_provider, valid_qasm2_no_meas):
    """Test emitting warning when number of qubits in the circuit exceeds the device capacity."""
    quera_device = mock_provider.get_device("quera_qasm_simulator")
    quera_device._target_spec = None
    quera_device.set_options(validate=1)

    with patch.object(
        type(quera_device), "num_qubits", new_callable=PropertyMock
    ) as mock_num_qubits:
        mock_num_qubits.return_value = 1

        with pytest.warns(
            UserWarning, match=r"Number of qubits in the circuit \(2\) exceeds.*capacity \(1\)"
        ):
            quera_device.validate([valid_qasm2_no_meas])


def test_device_validate_emit_warning_for_target_spec_validator(mock_provider, valid_qasm2):
    """Test emitting warning when target spec validator raises an exception."""
    quera_device = mock_provider.get_device("quera_qasm_simulator")
    quera_device.set_options(validate=1)

    msg = (
        r"OpenQASM programs submitted to the quera_qasm_simulator cannot contain measurement gates"
    )
    with pytest.warns(UserWarning, match=msg):
        quera_device.validate([valid_qasm2])


def test_resolve_noise_model_raises_for_unsupported(mock_quera_device):
    """Test raising exception when no noise models are supported by device."""
    assert mock_quera_device.profile.noise_models is None

    with pytest.raises(ValueError) as excinfo:
        mock_quera_device._resolve_noise_model("ideal")
    assert "Noise models are not supported by this device." in str(excinfo)
